using UnityEngine;
using UnityEngine.Rendering;
using UnityEngine.PostProcessing;
using System;
using System.Collections;
using MVR.FileManagementSecure;

namespace moyashi {
  class RenderTextureSwapper : MonoBehaviour {
    RenderTexture originalRT;
    RenderTexture originalPreviewRT;
    RenderTexture replacementRT;
    RenderTexture replacementPreviewRT;
    // bool on = false;

    public CommandBufferResources commandBufferResources;

    CommandBuffer hiResCommandBuffer;
    CommandBuffer previewCommandBuffer;

    Camera screenshotPreviewCamera;

    Renderer previewRenderer;

    private IEnumerator coroutine;
    private bool isRenderingOneFrame = false;
    private int renderFrames = 0;

    public void setPreviewResolution(CommandBufferResources resources, Vector2 res) {
      var preview = SuperController.singleton.hiResScreenshotPreview;

      if (replacementPreviewRT != null) {
        if (previewRenderer != null) {
          screenshotPreviewCamera.targetTexture = originalPreviewRT;
          previewRenderer.sharedMaterial.mainTexture = originalPreviewRT;
        }

        RenderTexture.ReleaseTemporary(replacementPreviewRT);
      }

      RenderTextureDescriptor desc = new RenderTextureDescriptor((int)res.x, (int)res.y, RenderTextureFormat.ARGB32, 24);
      desc.msaaSamples = 2;
      desc.autoGenerateMips = false;
      desc.useMipMap = false;
      desc.sRGB = true;
      replacementPreviewRT = RenderTexture.GetTemporary(desc);
      replacementPreviewRT.filterMode = FilterMode.Trilinear;
      replacementPreviewRT.wrapMode = TextureWrapMode.Clamp;

      if (previewRenderer != null) {
        screenshotPreviewCamera.targetTexture = replacementPreviewRT;
        previewRenderer.sharedMaterial.mainTexture = replacementPreviewRT;
      }

      // change aspect ratio of the preview panel
      Transform renderPlane = preview.transform.Find("Render");
      float aspectRatio = res.x / res.y;
      float defaultAspectRatio = 16.0f / 9.0f;
      renderPlane.localScale = new Vector3(aspectRatio / defaultAspectRatio, 1.0f, 1.0f);

      RectTransform backCanvas = preview.transform.Find("BackCanvas") as RectTransform;
      const float padding = 10.0f;
      const float height = 150.0f;
      backCanvas.offsetMax = new Vector2(aspectRatio * height + padding, height + padding);
      backCanvas.offsetMin = -backCanvas.offsetMax;

      if (previewCommandBuffer != null) {
        screenshotPreviewCamera.RemoveCommandBuffer(EffectsManager.POST_PROC_EVENT, previewCommandBuffer);
      }

      {
        var cam = screenshotPreviewCamera;
        var camWidth = cam.pixelWidth;
        var camHeight = cam.pixelHeight;

        previewCommandBuffer = CommandBufferBuilder.buildFullCommandBuffer(resources, camWidth, camHeight, false);
        cam.AddCommandBuffer(EffectsManager.POST_PROC_EVENT, previewCommandBuffer);
      }
    }

    public void setRenderResolution(CommandBufferResources resources, Vector2 res) {
      var cam = SuperController.singleton.hiResScreenshotCamera;

      if (replacementRT != null) {
        cam.targetTexture = null;
        RenderTexture.ReleaseTemporary(replacementRT);
      }

      RenderTextureDescriptor desc = new RenderTextureDescriptor((int)res.x, (int)res.y, RenderTextureFormat.ARGB32, 24);
      desc.msaaSamples = 8;
      desc.autoGenerateMips = false;
      desc.useMipMap = false;
      desc.sRGB = true;
      replacementRT = RenderTexture.GetTemporary(desc);
      replacementRT.filterMode = FilterMode.Trilinear;
      replacementRT.wrapMode = TextureWrapMode.Clamp;

      // cam.targetTexture = replacementRT;
      cam.targetTexture = null;

      // if (previewRenderer != null) {
      //   previewRenderer.sharedMaterial.mainTexture = replacementRT;
      // }

      // commandBufferResources = resources;
    }

    // i think the way supershot is working is:
    // - set targetTexture to null on screenshot camera so it errors when trying to save a screenshot
    // - create replacement for screenshot camera
    // - create own callbacks etc for replacement screenshot camera to write to disk
    // - starts a coroutine that waits until user presses (or releases?) (!sc.GetRightSelect() && !sc.GetLeftSelect() && !sc.GetMouseSelect())
    // - manually calls prerender(?) on the replacement cam
    // SEEMS ANNOYINGLY COMPLICATED

    // GameObject cameraContainer = GameObject.Find("Moyashi_PostProcessingCamera");
    // if (cameraContainer == null) {
    //   cameraContainer = new GameObject("Moyashi_PostProcessingCamera");
    // }

    // postProcessingCamera = cameraContainer.GetComponent<Camera>();

    // if (postProcessingCamera == null) {
    //   postProcessingCamera = cameraContainer.AddComponent<Camera>();
    // }
    
    // // TODO: is this necessary
    // postProcessingCamera.CopyFrom(mainCamera);

    // // works perfectly in vr (in conjunction with setting culling matrix)
    // if (postProcessingCamera.stereoEnabled) {
    //   postProcessingCamera.transform.parent = SuperController.singleton.heightAdjustTransform;
    // } else {
    //   postProcessingCamera.transform.parent = mainCamera.transform;
    // }



    private void OnEnable() {
      var cam = SuperController.singleton.hiResScreenshotCamera;
      var preview = SuperController.singleton.hiResScreenshotPreview;

      cam.enabled = false;

      originalRT = cam.targetTexture;
      originalPreviewRT = cam.targetTexture; // ???

      GameObject cameraContainer = GameObject.Find("Moyashi_ScreenshotPreviewCamera");
      if (cameraContainer == null) {
        cameraContainer = new GameObject("Moyashi_ScreenshotPreviewCamera");
      }

      screenshotPreviewCamera = cameraContainer.GetComponent<Camera>();

      if (screenshotPreviewCamera == null) {
        screenshotPreviewCamera = cameraContainer.AddComponent<Camera>();
      }

      screenshotPreviewCamera.CopyFrom(SuperController.singleton.hiResScreenshotCamera);
      screenshotPreviewCamera.transform.parent = SuperController.singleton.hiResScreenshotCamera.transform;
      screenshotPreviewCamera.enabled = false;

      // this bit working
      // RenderTextureDescriptor desc = new RenderTextureDescriptor(1024, 1024, RenderTextureFormat.ARGB32, 24);
      // desc.msaaSamples = 1;
      // desc.autoGenerateMips = true;
      // desc.useMipMap = true;
      // desc.sRGB = true;
      // replacementRT = RenderTexture.GetTemporary(desc);
      // replacementRT.filterMode = FilterMode.Trilinear;
      // replacementRT.wrapMode = TextureWrapMode.Clamp;

      // cam.targetTexture = replacementRT;

      if (preview) {
        SuperController.LogMessage("preview present");
        var render = preview.Find("Render");

        if (render) {
          previewRenderer = render.GetComponent<Renderer>();
          // originalPreviewRT = previewRenderer.sharedMaterial.mainTexture; // this aint rite
          // previewRenderer.sharedMaterial.mainTexture = replacementRT;
        }
      }

      // add post processing profile to hires and preview cameras
      var previewPostProcessingBehaviour = screenshotPreviewCamera.GetComponent<PostProcessingBehaviour>();
      if (previewPostProcessingBehaviour == null) {
        previewPostProcessingBehaviour = screenshotPreviewCamera.gameObject.AddComponent<PostProcessingBehaviour>();
      } 

      var hiResPostProcessingBehaviour = cam.GetComponent<PostProcessingBehaviour>();
      if (hiResPostProcessingBehaviour == null) {
        hiResPostProcessingBehaviour = cam.gameObject.AddComponent<PostProcessingBehaviour>();
      }

      // WHOA BAD PROGRAMMING
      previewPostProcessingBehaviour.profile = EffectsManager.profile;
      hiResPostProcessingBehaviour.profile = EffectsManager.profile;

      coroutine = waitForScreenshotIntent();
      StartCoroutine(coroutine);

      SuperController.LogMessage("on enable");
      // on = true;
    }

    private void OnDisable() {
      SuperController.LogMessage("ScreenshotManager OnDisable start");

      if (previewRenderer) {
        previewRenderer.sharedMaterial.mainTexture = originalRT;
      }
      StopCoroutine(coroutine);
      SuperController.singleton.hiResScreenshotCamera.targetTexture = originalRT;

      RenderTexture.ReleaseTemporary(replacementRT);
      RenderTexture.ReleaseTemporary(replacementPreviewRT);

      // put preview back to the way it was
      var preview = SuperController.singleton.hiResScreenshotPreview;
      Transform renderPlane = preview.transform.Find("Render");
      renderPlane.localScale = new Vector3(1.0f, 1.0f, 1.0f);

      RectTransform backCanvas = preview.transform.Find("BackCanvas") as RectTransform;
      const float padding = 10.0f;
      const float height = 150.0f;
      float defaultAspectRatio = 16.0f / 9.0f;
      backCanvas.offsetMax = new Vector2(defaultAspectRatio * height + padding, height + padding);
      backCanvas.offsetMin = -backCanvas.offsetMax;

      GameObject cameraContainer = GameObject.Find("Moyashi_ScreenshotPreviewCamera");
      DestroyImmediate(cameraContainer);

      // renable the built in hi res camera
      SuperController.singleton.hiResScreenshotCamera.enabled = true;

      SuperController.LogMessage("ScreenshotManager OnDisable");
    }

    private void OnPreRender() {
      SuperController.LogMessage("on pre render");
      var cam = SuperController.singleton.hiResScreenshotCamera;

      // RenderTextureDescriptor desc = new RenderTextureDescriptor(1024, 1024, RenderTextureFormat.ARGB32, 24);
      // replacementRT = RenderTexture.GetTemporary(desc);

      // cam.targetTexture = replacementRT;
      // cam.enabled = isRenderingOneFrame;

      // {
      //   var camWidth = cam.targetTexture.width;
      //   var camHeight = cam.targetTexture.height;

      //   hiResCommandBuffer = CommandBufferBuilder.buildFullCommandBuffer(commandBufferResources, camWidth, camHeight, false);
      //   cam.AddCommandBuffer(CameraEvent.BeforeImageEffects, hiResCommandBuffer);

      //   SuperController.LogMessage("added command buffer");

      //   isRenderingOneFrame = true;
        
      // }

      // if (isRenderingOneFrame) {
      //   UnityEngine.RenderTexture.active = replacementRT;
      // }

      if (renderFrames == 2) {
        var camWidth = cam.targetTexture.width;
        var camHeight = cam.targetTexture.height;

        hiResCommandBuffer = CommandBufferBuilder.buildFullCommandBuffer(commandBufferResources, camWidth, camHeight, false);
        cam.AddCommandBuffer(CameraEvent.BeforeImageEffects, hiResCommandBuffer);

        SuperController.LogMessage("added command buffer");
      }

      if (renderFrames <= 0) {
        cam.enabled = false;
      }
    }

    private void OnPostRender() {
      // SuperController.LogMessage("on post render");
      // SuperController.singleton.hiResScreenshotCamera.targetTexture = previewRT;
      // RenderTexture.ReleaseTemporary(replacementRT);
      // SuperController.LogMessage("on post render done");

      if (renderFrames > 0) {
        renderFrames--;        
      }

      if (renderFrames == 1) {
        StartCoroutine(stopCamera());
      }
    }

    public void LateUpdate() {
      if (isRenderingOneFrame) {
        // StartCoroutine(saveScreenshot());
      }
    }

    private void Update() {
      screenshotPreviewCamera.enabled = SuperController.singleton.hiResScreenshotPreview.gameObject.activeSelf;
      screenshotPreviewCamera.fieldOfView = SuperController.singleton.hiResScreenshotCamera.fieldOfView;
    }

    private IEnumerator stopCamera() {
      yield return new WaitForEndOfFrame();
      var cam = SuperController.singleton.hiResScreenshotCamera;
      cam.enabled = false;
      cam.targetTexture = null;
      cam.RemoveCommandBuffer(CameraEvent.BeforeImageEffects, hiResCommandBuffer);
      isRenderingOneFrame = false;

      StartCoroutine(saveScreenshot());
    }

    private IEnumerator saveScreenshot() {
      yield return new WaitForEndOfFrame();

      var cam = SuperController.singleton.hiResScreenshotCamera;
      Texture2D tex = null;        

      // This is pulled from SuperController.ProcessHiResScreenshot because everything is
      // private/protected can't just call it manually.
      try {        
        UnityEngine.RenderTexture.active = replacementRT;
        var width = replacementRT.width;
        var height = replacementRT.height;
        tex = new Texture2D(width, height,(TextureFormat)5, false);
        tex.ReadPixels(new Rect(0.0f, 0.0f, width, height), 0, 0);
        tex.Apply(); // ???

        byte[] bytes = ImageConversion.EncodeToJPG(tex, 100);
        int num = (int)(DateTime.UtcNow - new System.DateTime(1970, 1, 1)).TotalSeconds;

        string filename = $"{num}.jpg";
        string path = FileManagerSecure.NormalizePath("Saves/screenshots/");

        SuperController.LogMessage($"saving to {path + filename}");
        FileManagerSecure.WriteAllBytes(path + filename, bytes);

        if (SkyshopLightController.singleton != null) {
          SkyshopLightController.singleton.Flash();
        }
      } finally {
        if (tex != null) {
          Destroy(tex);
        }
        
        // isRenderingOneFrame = false;
        // cam.enabled = false;
        // cam.targetTexture = null;
        // cam.RemoveCommandBuffer(CameraEvent.BeforeImageEffects, hiResCommandBuffer);

        // TODO: trigger GC manually here?

      }



      // yield return null;
    }

    private IEnumerator waitForScreenshotIntent() {
      var screenshotIntentMonitor = new ScreenshotIntentMonitor();
      while (true) {
        SuperController.LogMessage("waiting for screenshot intent");
        yield return StartCoroutine(screenshotIntentMonitor);

        if (SuperController.singleton.hiResScreenshotPreview.gameObject.activeSelf) {
          var cam = SuperController.singleton.hiResScreenshotCamera;
          cam.targetTexture = replacementRT;
          // cam.forceIntoRenderTexture = true;
          // cam.cullingMask = screenshotPreviewCamera.cullingMask;
          // cam.cullingMatrix = screenshotPreviewCamera.cullingMatrix;
          // cam.Render();
          cam.enabled = true;
          renderFrames = 2;
          SuperController.LogMessage("taking screenshot");
        }
        
        // TODO: probably don't need this if we can somehow wait for Render above
        yield return new WaitForSeconds(0.25f);
      }
    }
  }

  class ScreenshotIntentMonitor : CustomYieldInstruction {
    public override bool keepWaiting { 
      get { 
        return (
          !SuperController.singleton.GetRightSelect() && 
          !SuperController.singleton.GetLeftSelect() && 
          !SuperController.singleton.GetMouseSelect()
        ); 
      }
    }
  }

  class ScreenshotManager {
    public static void setHiResScreenshotCameraResolution(CommandBufferResources resources, Vector2 renderRes, Vector2 previewRes) {
      var cam = SuperController.singleton.hiResScreenshotCamera;
      SuperController.LogMessage($"{cam.targetTexture}");

      RenderTextureSwapper swapper = cam.GetComponent<RenderTextureSwapper>();

      // swapper gets cleaned up in PostProcessing OnDestroy callback so dont need to worry about it here

      if (swapper == null) {
        SuperController.LogMessage("swapper was null, adding one to hiResScreenshotCamera");
        swapper = cam.gameObject.AddComponent<RenderTextureSwapper>();
      }

      swapper.commandBufferResources = resources;

      SuperController.LogMessage($"render {renderRes}");
      SuperController.LogMessage($"preview {previewRes}");

      swapper.setRenderResolution(resources, renderRes);
      swapper.setPreviewResolution(resources, previewRes);
    }
  }
}